"""Package teneva, module core.stat: helper functions for processing statistics.

This module contains the helper functions for processing statistics, including
computation of the CDF function and its confidence bounds.

"""
import numpy as np


def cdf_confidence(F, alpha=0.05):
    """Constructs a Dvoretzky-Kiefer-Wolfowitz confidence band for the CDF.

    Args:
        F (np.ndarray): the empirical distributions in the form of 1D np.ndarray
            of length m.
        alpha (float): alpha for a (1 - alpha) confidence band.

    Returns:
        np.ndarray: the CDF lower bound in the form of 1D np.ndarray
            of length m.
        np.ndarray: the CDF upper bound in the form of 1D np.ndarray
            of length m.

    Note:
        The description of this algorithm is presented in the work: Wasserman
        L., "All of Nonparametric Statistics".

    """
    eps = np.sqrt(np.log(2. / alpha) / (2 * len(F)))
    return np.clip(F - eps, 0, 1), np.clip(F + eps, 0, 1)


def cdf_getter(x):
    """Build the getter for CDF.

    Args:
        x (np.ndarray): 1D points in the form of np.ndarray or list.

    Returns:
        function: the function that computes CDF values (input is point in the
            form of np.ndarray and output is the corresponding float CDF value).

    """
    x = np.array(x, copy=True)
    x.sort()
    y = np.linspace(1./len(x), 1, len(x))

    x = np.r_[-np.inf, x]
    y = np.r_[0, y]

    def cdf(z):
        return y[np.searchsorted(x, z, 'right') - 1]

    return cdf



"""Package teneva, module core.cheb: Chebyshev interpolation in the TT-format.

This module contains the functions for construction of the Chebyshev
interpolation in the TT-format as well as calculating the values of the
function using the constructed interpolation coefficients.

"""
import numpy as np


from .cross import cross
from .grid import grid_prep_opts
from .grid import ind2poi
from .tensor import copy
from .tensor import shape
from .tensor import truncate


def cheb_bld(f, a, b, n, **args):
    """Compute the function values on the Chebyshev grid.

    Args:
        f (function): function f(X) for interpolation, where X should be 2D
            np.ndarray of the shape [samples, dimensions]. The function should
            return 1D np.ndarray of the length equals to samples.
        a (list): grid lower bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the lower bounds for each
            dimension will be the same.
        b (list): grid upper bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the upper bounds for each
            dimension will be the same.
        n (list): tensor size for each dimension (list or np.ndarray of length
            "d"). It may be also float, then the size for each dimension will be
            the same.
        args (dict): named arguments for TT-CAM ("cross") function except the
            target function "f", i.e. "(Y0, e, evals, nswp, dr_min, dr_max,
            cache, info)". Note that initial approximation "Y0" and accuracy
            "e" are required.

    Returns:
        Y (list): TT-Tensor with function values on the Chebyshev grid.

    Note:
        At least one of the variables "a", "b", "n" must be a list (to be able
        to automatically determine the dimension).

    """
    a, b, n = grid_prep_opts(a, b, n)
    Y = cross(lambda I: f(ind2poi(I.astype(int), a, b, n, 'cheb')), **args)
    return Y


def cheb_get(X, A, a, b, z=0.):
    """Compute the Chebyshev approximation in given points (approx. f(X)).

    Args:
        X (np.ndarray): spatial points of interest (it is 2D array of the shape
            [samples, d], where "d" is the number of dimensions).
        A (list): TT-tensor of the interpolation coefficients (it has d
            dimensions).
        a (list): grid lower bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the lower bounds for each
            dimension will be the same.
        b (list): grid upper bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the upper bounds for each
            dimension will be the same.
        z (float): the value for points, which are outside the spatial grid.

    Returns:
        np.ndarray: approximated function values in X points (it is 1D array of
            the shape [samples]).

    """
    d = len(A)
    n = shape(A)
    m = X.shape[0]
    a, b, n = grid_prep_opts(a, b, n, d)

    # TODO: check if this operation is effective. It may be more profitable to
    # generate polynomials for each tensor mode separately:
    T = cheb_pol(X, a, b, max(n))

    Y = np.ones(m) * z
    for i in range(m):
        if np.max(a - X[i, :]) > 1.E-16 or np.max(X[i, :] - b) > 1.E-16:
            # We skip the points outside the grid bounds:
            continue

        Q = np.einsum('rkq,k->rq', A[0], T[:n[0], i, 0])
        for j in range(1, d):
            Q = Q @ np.einsum('rjq,j->rq', A[j], T[:n[j], i, j])
        Y[i] = Q[0, 0]

    return Y


def cheb_get_full(A, a, b, m=None, e=1.E-6):
    """Compute the Chebyshev approximation (TT-tensor) on the full given grid.

    Args:
        A (list): TT-tensor of the interpolation coefficients (it has d
            dimensions).
        a (float): grid lower bounds for each dimension (the lower bounds for
            each dimension should be the same).
        b (float): grid upper bounds for each dimension (the upper bounds for
            each dimension should be the same).
        m (int): number of grid points for each dimension (>= 2). If is not set,
            then original grid size (from the interpolation) will be used.
        e (float): accuracy for truncation of the result (> 0).

    Returns:
        list: TT-tensor of the approximated function values on the full grid. (m x m x ... x m).

    Note:
        This function works correctly only for grids with an equal number of
        points for each mode!

    """
    d = len(A)
    n = A[0].shape[1]
    m = m or n
    I = np.arange(m).reshape((1, -1))
    X = ind2poi(I, a, b, m, 'cheb').reshape(-1)
    T = cheb_pol(X, a, b, n)
    Q = []
    for i in range(d):
        Q.append(np.einsum('riq,ij->rjq', A[i], T))

    return truncate(Q, e)


def cheb_int(Y, e=1.E-6):
    """Compute the TT-tensor for Chebyshev interpolation coefficients.

    Args:
        Y (list): TT-tensor with function values on the Chebyshev grid.
        e (float): accuracy for truncation of the result (> 0).

    Returns:
        list: TT-tensor that collects interpolation coefficients. It has the
            same shape as the given tensor Y.

    """
    A = copy(Y)
    for k in range(len(A)):
        r, m, q = A[k].shape
        A[k] = np.swapaxes(A[k], 0, 1)
        A[k] = A[k].reshape((m, -1))
        A[k] = np.vstack([A[k], A[k][m-2 : 0 : -1, :]])
        A[k] = np.fft.fft(A[k], axis=0).real
        A[k] = A[k][:m, :] / (m - 1)
        A[k][0, :] /= 2.
        A[k][m-1, :] /= 2.
        A[k] = A[k].reshape((m, r, q))
        A[k] = np.swapaxes(A[k], 0, 1)
    return truncate(A, e)


def cheb_pol(X, a, b, m):
    """Compute the Chebyshev polynomials in the given points.

    Args:
        X (np.ndarray): spatial points of interest (it is 2D array of the shape
            [samples, d], where d is a number of dimensions).
        a (list): grid lower bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the lower bounds for each
            dimension will be the same.
        b (list): grid upper bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the upper bounds for each
            dimension will be the same.
        m (int): maximum order for Chebyshev polynomial (>= 1). The polynomials
            of the order 0,1,...,m-1 will be computed.

    Returns:
        np.ndarray: values of the Chebyshev polynomials of the order 0,1,...,m-1
            in X points (it is 3D array of the shape [m x samples x d]).

    Note:
        Before calculating polynomials, the points are scaled from [a, b] to standard [-1, 1] limits.

    """
    d = X.shape[-1]
    reps = X.shape[0] if len(X.shape) > 1 else None
    a, b = grid_prep_opts(a, b, None, d, reps)
    X = (2. * X - b - a) / (b - a)

    T = np.ones([m] + list(X.shape))

    if m < 2:
        return T

    T[1, ] = X.copy()
    for k in range(2, m):
        T[k, ] = 2. * X * T[k - 1, ] - T[k - 2, ]

    return T


def cheb_sum(A, a, b):
    """Integrate the function from its Chebyshev approximation in the TT-format.

    Args:
        A (list): TT-tensor of the interpolation coefficients (it has d
            dimensions).
        a (list): grid lower bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the lower bounds for each
            dimension will be the same.
        b (list): grid upper bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the upper bounds for each
            dimension will be the same.

    Returns:
        float: the value of the integral.

    Note:
        This function works only for symmetric grids!

    """
    d = len(A)
    a, b = grid_prep_opts(a, b, None, d)

    Y = copy(A)
    v = np.array([[1.]])
    for k in range(len(Y)):
        r, m, q = Y[k].shape
        Y[k] = np.swapaxes(Y[k], 0, 1)
        Y[k] = Y[k].reshape(m, -1)
        p = np.arange(Y[k].shape[0])[::2]
        p = np.repeat(p.reshape(-1, 1), Y[k].shape[1], axis=1)
        Y[k] = np.sum(Y[k][::2, :] * 2. / (1. - p**2), axis=0)
        Y[k] = Y[k].reshape(r, q)
        v = (v @ Y[k]) * (b[k] - a[k]) / 2.

    return v[0, 0]



"""Package teneva, module demo.demo_func: base class for benchmarks."""
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.ticker import LinearLocator, FormatStrFormatter
import numpy as np


import teneva


class DemoFunc:
    def __init__(self, d, name='Demo'):
        """Demo function.

        Args:
            d (int): number of dimensions.
            name (str): display name of the function.

        """
        self.d = d
        self.name = name

        self.m_trn = 0
        self.m_tst = 0

        self.set_grid()
        self.set_lim(-1., +1.)
        self.set_min(None, None)
        self.set_perm()

    def build_trn(self, m, is_ind=False):
        """Generate train dataset from random indices or spatial points.

        Args:
            m (int or float): number of points to generate.
            is_ind (bool): if True, then grid indices will be generated
                (I_trn), otherwise the spatial points (X_trn) will be generated.

        Note:
            In case is_ind = False, class instance variables X_trn (spatial
                points) and Y_trn (function values) will be saved.

            In case is_ind = True, class instance variables I_trn (grid
                indices), X_trn (spatial points related to grid indices) and
                Y_trn (function values) will be saved. Indices I_trn will be
                generated from LHS distribution.

        """
        self.m_trn = int(m)
        a, b, n = self.a, self.b, self.n

        if is_ind:
            func = teneva.sample_lhs
            self.I_trn = func(self.n, self.m_trn)
            self.X_trn = teneva.ind2poi(self.I_trn, a, b, n, self.kind)
            self.Y_trn = self.comp(self.X_trn)
        else:
            func = np.random.uniform
            self.I_trn = None
            self.X_trn = np.vstack([func(a, b) for _ in range(self.m_trn)])
            self.Y_trn = self.comp(self.X_trn)

    def build_tst(self, m, is_ind=False):
        """Generate test dataset from random indices or spatial points.

        Args:
            m (int or float): number of points to generate.
            is_grid (bool): if True, then grid indices will be generated
                (I_trn), otherwise the spatial points (X_trn) will be generated.

        Note:
            In case is_ind = False, class instance variables X_tst (spatial
                points) and Y_tst (function values) will be saved.

            In case is_ind = True, class instance variables I_tst (grid
                indices), X_tst (spatial points related to grid indices) and
                Y_tst (function values) will be saved. Indices I_trn will be
                generated from uniform random choice of indices.

        """
        self.m_tst = int(m)
        a, b, n = self.a, self.b, self.n

        if is_ind:
            func = np.random.choice
            self.I_tst = np.vstack([func(k, self.m_tst) for k in n]).T
            self.X_tst = teneva.ind2poi(self.I_tst, a, b, n, self.kind)
            self.Y_tst = self.comp(self.X_tst)
        else:
            func = np.random.uniform
            self.I_tst = None
            self.X_tst = np.vstack([func(a, b) for _ in range(self.m_tst)])
            self.Y_tst = self.comp(self.X_tst)

    def calc(self, x):
        """Calculate the function in the given point.

        Args:
            x (np.ndarray): point (function input) in the form of array of the
                shape [d], where "d" is the dimension of the input.

        Returns:
            float: the value of the function in given point.

        """
        z = x if self.perm is None else x[self.perm]
        return self._calc(z)

    def calc_grid(self, i):
        """Calculate the function in the given grid index.

        Args:
            i (np.ndarray): grid index in the form of array of the shape
                [d], where "d" is the dimension of the input.

        Returns:
            float: the function value in the point related to given grid index.

        Note:
            Grid params may be set using "set_grid" function.

        """
        x = teneva.ind2poi(i, self.a, self.b, self.n, self.kind)
        return self.calc(x)

    def check_trn(self, Y):
        """Compute the error of TT-tensor on the dataset.

        Args:
            Y (list): TT-tensor.

        Returns:
            float: the relative difference between two lists of values, i.e.,
                norm(Y - Y_trn) / norm(Y_trn).

        """
        if not self.m_trn:
            # raise ValueError('Train points are not prepared')
            return -1

        if self.I_trn is None:
            # Train points are spatial:
            if self.kind != 'cheb':
                raise ValueError('Can check only "cheb" spatial train points')
            A = teneva.cheb_int(Y)
            Z = teneva.cheb_get(self.X_trn, A, self.a, self.b)
        else:
            # Train points are indices:
            get = teneva.getter(Y)
            Z = np.array([get(i) for i in self.I_trn])

        return np.linalg.norm(Z - self.Y_trn) / np.linalg.norm(self.Y_trn)

    def check_tst(self, Y):
        """Compute the error of TT-tensor on the test dataset.

        Args:
            Y (list): TT-tensor.

        Returns:
            float: the relative difference between two lists of values, i.e.,
                norm(Y - Y_tst) / norm(Y_tst).

        """
        if not self.m_tst:
            # raise ValueError('Test points are not prepared')
            return -1

        if self.I_tst is None:
            # Test points are spatial:
            if self.kind != 'cheb':
                raise ValueError('Can check only "cheb" spatial test points')
            A = teneva.cheb_int(Y)
            Z = teneva.cheb_get(self.X_tst, A, self.a, self.b)
        else:
            # Test points are indices:
            get = teneva.getter(Y)
            Z = np.array([get(i) for i in self.I_tst])

        return np.linalg.norm(Z - self.Y_tst) / np.linalg.norm(self.Y_tst)

    def comp(self, X):
        """Compute the function in the given points.

        Args:
            X (np.ndarray): points (function inputs) in the form of array of the
                shape [samples, d], where "samples" is the number of samples and
                "d" is the dimension of the input.

        Returns:
            np.ndarray: the values of the function in given points in the form
                of array of the shape [samples].

        """
        Z = X if self.perm is None else X[self.perm]
        return self._comp(X)

    def comp_grid(self, I):
        """Compute the function in the given grid indices.

        Args:
            I (np.ndarray): grid indices in the form of array of the shape
                [samples, d], where "samples" is the number of samples and "d"
                is the dimension of the input.

        Returns:
            np.ndarray: the values of the function in the points related to
                given grid indices in the form of array of the shape [samples].

        Note:
            Grid params may be set using "set_grid" function.

        """
        X = teneva.ind2poi(I, self.a, self.b, self.n, self.kind)
        return self.comp(X)

    def plot(self, k=1000):
        """Plot the function for the 2D case.

        Args:
            k (int): number of points for each dimension.

        """
        if self.d != 2:
            raise ValueError('Plot is supported only for 2D case')

        X1 = np.linspace(self.a[0], self.b[0], k)
        X2 = np.linspace(self.a[1], self.b[1], k)
        X1, X2 = np.meshgrid(X1, X2)
        X = np.hstack([X1.reshape(-1, 1), X2.reshape(-1, 1)])

        Y = self.comp(X)
        Y = Y.reshape(X1.shape)

        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca(projection='3d')
        ax.set_title(self.name + ' function')
        surf = ax.plot_surface(X1, X2, Y, cmap=cm.coolwarm,
            linewidth=0, antialiased=False)
        ax.zaxis.set_major_locator(LinearLocator(10))
        ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))
        fig.colorbar(surf, shrink=0.3, aspect=10)
        plt.show()

    def set_grid(self, n=10, kind='uni'):
        """Set grid options for function discretization.

        Args:
            n (list): grid size for each dimension (list or np.ndarray of length
                "d"). It may be also float, then the size for each dimension
                will be the same.
            kind (str): the grid kind, it may be "uni" (uniform grid) and "cheb"
                (Chebyshev grid).

        """
        if isinstance(n, (int, float)):
            n = [int(n)] * self.d
        if isinstance(n, list):
            n = np.array(n, dtype=int)
        self.n = n

        if not kind in ['uni', 'cheb']:
            raise ValueError(f'Unknown grid type "{kind}"')
        self.kind = kind

    def set_lim(self, a, b):
        """Set spatial grid bounds.

        Args:
            a (list): grid lower bounds for each dimension (list or np.ndarray
                of length "d"). It may be also float, then the lower bounds for each dimension will be the same.
            b (list): grid upper bounds for each dimension (list or np.ndarray
                of length "d"). It may be also float, then the upper bounds for
                each dimension will be the same.

        """
        if isinstance(a, (int, float)):
            a = [a] * self.d
        if isinstance(a, list):
            a = np.array(a, dtype=float)
        self.a = a

        if isinstance(b, (int, float)):
            b = [b] * self.d
        if isinstance(b, list):
            b = np.array(b, dtype=float)
        self.b = b

    def set_min(self, x_min, y_min):
        """Set the exact global minimum of the function.

        Args:
            x_min (np.ndarray): argument of the function that provides its
                minimum value (list or np.ndarray of length "d").
            y_min (float): minimum value of the function.

        """
        if isinstance(x_min, list):
            x_min = np.array(x_min, dtype=float)
        self.x_min = x_min

        self.y_min = y_min

    def set_perm(self, perm=None):
        """Set the permutation of the function arguments.

        Args:
            perm (list, np.ndarray): new ordering of indices.

        """
        self.perm = perm
        if self.perm is None:
            return

        self.a = self.a[self.perm]
        self.b = self.b[self.perm]
        self.n = self.n[self.perm]

        if self.m_trn and self.I_trn is not None:
            self.I_trn = self.I_trn[:, self.perm]
        if self.m_trn and self.X_trn is not None:
            self.X_trn = self.X_trn[:, self.perm]

        if self.m_tst and self.I_tst is not None:
            self.I_tst = self.I_tst[:, self.perm]
        if self.m_tst and self.X_tst is not None:
            self.X_tst = self.X_tst[:, self.perm]

    def _calc(self, x):
        raise NotImplementedError()

    def _comp(self, X):
        raise NotImplementedError()



"""Package teneva, module core.cross: cross approximation in the TT-format.

This module contains the function "cross" which computes the TT-approximation
for implicit tensor given functionally by multidimensional cross approximation
method in the TT-format (TT-CAM).

"""
import numpy as np


from .grid import ind2str
from .grid import str2ind
from .maxvol import maxvol
from .maxvol import maxvol_rect
from .tensor import accuracy
from .tensor import copy
from .tensor import truncate


def cross(f, Y0, e, evals=None, nswp=10, dr_min=1, dr_max=2, cache=None,
    info={}):
    """Compute the TT-approximation for implicit tensor given functionally.

    Args:
        f (function): function f(I) which computes tensor elements for a given
            set of multi-indices I, where I is 2D np.ndarray of the shape
            [samples x dimensions]. The function should return 1D np.ndarray of
            the length equals to samples.
        Y0 (list): TT-tensor, which is the initial approximation for algorithm.
        e (float): accuracy (> 0) for truncation of the final result and
            algorithm convergence criterion. If between iterations the relative
            rate of solution change is less than this value, then the operation
            of the algorithm will be interrupted.
        evals (int): an optionally set limit on the maximum number of requests
            to the objective function. If specified, then the total number of
            requests will not exceed this value. Note that the actual number of
            requests may be less, since the values are requested in batches.
        nswp (int): maximum number of iterations (sweeps) of the algorithm.
        dr_min (int): minimum number of added rows in the process of adaptively
            increasing the TT-rank of the approximation using the algorithm
            maxvol_rect (see teneva.core.maxvol.maxvol_rect for more details).
            Note that "dr_min" should be no bigger than "dr_max".
        dr_max (int): maximum number of added rows in the process of adaptively
            increasing the TT-rank of the approximation using the algorithm
            maxvol_rect (see teneva.core.maxvol.maxvol_rect for more details).
            Note that "dr_max" should be no less than "dr_min".
        cache (dict): an optionally set dictionary, which will be filled with
            requested function values. Since the algorithm sometimes requests
            the same tensor indices, the use of such a cache in some cases can
            speed up the operation of the algorithm if the time to find a value
            in the cache is less than the time to calculate the function.
        info (dict): an optionally set dictionary, which will be filled with
            reference information about the process of the algorithm operation.
            At the end of the function work, it will contain parameters:
            "k_evals" - total number of requests to target function "f";
            "k_cache" - total number of requests to cache;
            "e" - the final value of the convergence criterion;
            "nswp" - the real number of performed iterations (sweeps);
            "stop" - stop type of the algorithm (see note below).

    Returns:
        list: TT-Tensor which approximates the implicit tensor.

    Note:
        Note that the end of the algorithm operation occurs when one of the
        three criteria is reached: 1) the maximum number of iterations ("nswp")
        performed; 2) the maximum allowable number of the objective function
        calls ("evals") has been done (more precisely, if the next request will
        result in exceeding this value, then algorithm will not perform this
        new request); 3) the convergence criterion ("e") is reached. The
        corresponding stop type ("nswp", "evals" or "e") will be written into
        the item "stop" of the "info" dictionary.

    """
    Y = copy(Y0)
    d = len(Y)

    Il = [np.empty((1, 0), dtype=int)] + [None for i in range(d)]
    Ir = [None for i in range(d)] + [np.empty((1, 0), dtype=int)]
    Ig = [_reshape(np.arange(G.shape[1], dtype=int), (-1, 1)) for G in Y]

    R = np.ones((1, 1))
    for i in range(d-1):
        G = np.tensordot(R, Y[i], 1)
        Y[i], Il[i+1], R = cross_prep_l2r(G, Ig[i], Il[i])

    R = np.ones((1, 1))
    for i in range(d-1, 0, -1):
        G = np.tensordot(Y[i], R, 1)
        Y[i], Ir[i], R = cross_prep_r2l(G, Ig[i], Ir[i+1])

    info['k_evals'] = 0
    info['k_cache'] = 0
    info['e'] = -1.
    info['nswp'] = 0
    info['stop'] = None

    def func(I):
        if cache is None:
            if evals is not None and info['k_evals'] + len(I) > evals:
                return None
            info['k_evals'] += len(I)
            return f(I)

        I_new = np.array([i for i in I if ind2str(i) not in cache])
        if len(I_new):
            if evals is not None and info['k_evals'] + len(I_new) > evals:
                return None
            Y_new = f(I_new)
            for k, i in enumerate(I_new):
                cache[ind2str(i)] = Y_new[k]

        info['k_cache'] += len(I) - len(I_new)
        info['k_evals'] += len(I_new)

        Y = np.array([cache[ind2str(i)] for i in I])
        return Y

    Yold = None
    for swp in range(nswp):
        R = np.ones((1, 1))
        for i in range(d):
            G = np.tensordot(R, Y[i], 1)
            y = func(cross_index_merge(Il[i], Ig[i], Ir[i+1]))
            if y is None:
                Y[i] = G
                info['stop'] = 'evals'
                return truncate(Y, e)
            Y[i], Il[i+1], R = cross_build_l2r(
                *G.shape, y, Ig[i], Il[i], dr_min, dr_max)
        Y[d-1] = np.tensordot(Y[d-1], R, 1)

        R = np.ones((1, 1))
        for i in range(d-1, -1, -1):
            G = np.tensordot(Y[i], R, 1)
            y = func(cross_index_merge(Il[i], Ig[i], Ir[i+1]))
            if y is None:
                Y[i] = G
                info['stop'] = 'evals'
                return truncate(Y, e)
            Y[i], Ir[i], R = cross_build_r2l(
                *G.shape, y, Ig[i], Ir[i+1], dr_min, dr_max)
        Y[0] = np.tensordot(R, Y[0], 1)

        info['nswp'] = swp + 1

        if Yold is not None:
            eps = accuracy(Y, Yold)
            info['e'] = eps

            if eps <= e:
                info['stop'] = 'e'
                return truncate(Y, e)

        Yold = copy(Y)

    info['stop'] = 'nswp'
    return truncate(Y, e)


def cross_build_l2r(r1, n, r2, y, Ig, I, dr_min, dr_max):
    G = _reshape(y, (-1, r2))
    Q, s, V = _svd(G)
    Ind, B = _maxvol_rect(Q, dr_min, dr_max)
    G = _reshape(B, (r1, n, -1))
    J = cross_index_stack_l2r(r1, Ig, I)[Ind, :]
    R = Q[Ind, :] @ np.diag(s) @ V
    return G, J, R


def cross_build_r2l(r1, n, r2, y, Ig, I, dr_min, dr_max):
    G = _reshape(y, (r1, -1)).T
    Q, s, V = _svd(G)
    Ind, B = _maxvol_rect(Q, dr_min, dr_max)
    G = _reshape(B.T, (-1, n, r2))
    J = cross_index_stack_r2l(r2, Ig, I)[Ind, :]
    R = (Q[Ind, :] @ np.diag(s) @ V).T
    return G, J, R


def cross_cache2data(cache):
    I = np.array([str2ind(s) for s in cache.keys()], dtype=int)
    Y = np.array([y for y in cache.values()])
    return I, Y


def cross_index_merge(i1, i2, i3):
    r1 = i1.shape[0] or 1
    r2 = i2.shape[0]
    r3 = i3.shape[0] or 1

    w1 = _kron(_ones(r3 * r2), i1)

    w2 = _kron(_kron(_ones(r3), i2), _ones(r1))

    if i3.size and r2:
        w3 = _kron(i3, _ones(r1 * r2))
    else:
        w3 = np.empty((w1.shape[0], 0))

    return np.hstack((w1, w2, w3))


def cross_index_stack_l2r(r, Ig, I):
    n = Ig.size
    J = _kron(Ig, _ones(r))
    if I.size:
        J = np.hstack((_kron(_ones(n), I), J))
    return _reshape(J, (n * r, -1))


def cross_index_stack_r2l(r, Ig, I):
    n = Ig.size
    J = _kron(_ones(r), Ig)
    if I.size:
        J = np.hstack((J, _kron(I, _ones(n))))
    return _reshape(J, (n * r, -1))


def cross_prep_l2r(G, Ig, I=None):
    r1, n, r2 = G.shape
    G = _reshape(G, (r1 * n, r2))

    Q, R = _qr(G)
    Ind, B = _maxvol(Q)

    G = _reshape(B.T, (-1, n, r2))
    J = cross_index_stack_l2r(r1, Ig, I)[Ind, :]
    R = Q[Ind, :] @ R
    return G, J, R


def cross_prep_r2l(G, Ig, I=None):
    r1, n, r2 = G.shape
    G = _reshape(G, (r1, n * r2)).T

    Q, R = _qr(G)
    Ind, B = _maxvol(Q)

    G = _reshape(B.T, (-1, n, r2))
    J = cross_index_stack_r2l(r2, Ig, I)[Ind, :]
    R = (Q[Ind, :] @ R).T
    return G, J, R


def _kron(a, b):
    return np.kron(a, b)


def _maxvol(a):
    return maxvol(a)


def _maxvol_rect(a, dr_min=1, dr_max=2, tau=1.1):
    N, u = a.shape

    if N <= u:
        I = np.arange(N, dtype=int)
        B = np.eye(N, dtype=a.dtype)
    else:
        I, B = maxvol_rect(a, tau, dr_min, dr_max)

    return I, B


def _ones(k, m=1):
    return np.ones((k, m), dtype=int)


def _qr(a):
    return np.linalg.qr(a)


def _reshape(a, shape):
    return np.reshape(a, shape, order='F')


def _svd(a, full_matrices=False):
    try:
        return np.linalg.svd(a, full_matrices=full_matrices)
    except:
        b = a + 1.E-14 * np.max(np.abs(a)) * np.random.randn(*a.shape)
        return np.linalg.svd(b, full_matrices=full_matrices)



"""Package teneva, module demo.demo_func_grienwank: function.

This module contains class that implements analytical Grienwank function
for demo and tests.

"""
import numpy as np


from .demo_func import DemoFunc


class DemoFuncGrienwank(DemoFunc):
    def __init__(self, d):
        """Grienwank function for demo and tests.

        Args:
            d (int): number of dimensions.

        Note:
            See https://www.sfu.ca/~ssurjano/griewank.html for details.

            See also Momin Jamil, Xin-She Yang. "A literature survey of
            benchmark functions for global optimization problems". Journal of
            Mathematical Modelling and Numerical Optimisation 2013; 4:150-194.

        """
        super().__init__(d, 'Grienwank')

        self.set_lim(-600., +600.)
        self.set_min([0.]*self.d, 0.)

    def _calc(self, x):
        y1 = np.sum(x**2) / 4000

        y2 = np.cos(x / np.sqrt(np.arange(self.d) + 1.))
        y2 = - np.prod(y2)

        y3 = 1.

        return y1 + y2 + y3

    def _comp(self, X):
        y1 = np.sum(X**2, axis=1) / 4000

        y2 = np.cos(X / np.sqrt(np.arange(self.d) + 1))
        y2 = - np.prod(y2, axis=1)

        y3 = 1.

        return y1 + y2 + y3



"""Package teneva, module demo.demo_func_ackley: function.

This module contains class that implements analytical Ackley function
for demo and tests.

"""
import numpy as np


from .demo_func import DemoFunc


class DemoFuncAckley(DemoFunc):
    def __init__(self, d, a=20., b=0.2, c=2.*np.pi):
        """Ackley function for demo and tests.

        Args:
            d (int): number of dimensions.
            a (float): parameter of the function.
            b (float): parameter of the function.
            c (float): parameter of the function.

        Note:
            See https://www.sfu.ca/~ssurjano/ackley.html for details.

            See also Momin Jamil, Xin-She Yang. "A literature survey of
            benchmark functions for global optimization problems". Journal of
            Mathematical Modelling and Numerical Optimisation 2013; 4:150-194.

        """
        super().__init__(d, 'Ackley')

        self.par_a = a
        self.par_b = b
        self.par_c = c

        self.set_lim(-32.768, +32.768)
        self.set_min([0.]*self.d, 0.)

    def _calc(self, x):
        y1 = np.sqrt(np.sum(x**2) / self.d)
        y1 = - self.par_a * np.exp(-self.par_b * y1)

        y2 = np.sum(np.cos(self.par_c * x))
        y2 = - np.exp(y2 / self.d)

        y3 = self.par_a + np.exp(1.)

        return y1 + y2 + y3

    def _comp(self, X):
        y1 = np.sqrt(np.sum(X**2, axis=1) / self.d)
        y1 = - self.par_a * np.exp(-self.par_b * y1)

        y2 = np.sum(np.cos(self.par_c * X), axis=1)
        y2 = - np.exp(y2 / self.d)

        y3 = self.par_a + np.exp(1.)

        return y1 + y2 + y3



from .demo_func import DemoFunc
from .demo_func_ackley import DemoFuncAckley
from .demo_func_brown import DemoFuncBrown
from .demo_func_grienwank import DemoFuncGrienwank
from .demo_func_michalewicz import DemoFuncMichalewicz
from .demo_func_piston import DemoFuncPiston
from .demo_func_rastrigin import DemoFuncRastrigin
from .demo_func_rosenbrock import DemoFuncRosenbrock
from .demo_func_schaffer import DemoFuncSchaffer
from .demo_func_schwefel import DemoFuncSchwefel



"""Package teneva, module demo.demo_func_michalewicz: function.

This module contains class that implements analytical Michalewicz function
for demo and tests.

"""
import numpy as np


from .demo_func import DemoFunc


class DemoFuncMichalewicz(DemoFunc):
    def __init__(self, d, m=10.):
        """Michalewicz function for demo and tests.

        Args:
            d (int): number of dimensions.
            m (float): parameter of the function.

        Note:
            See https://www.sfu.ca/~ssurjano/michal.html for details.

            See also Charlie Vanaret, Jean-Baptiste Gotteland, Nicolas Durand,
            Jean-Marc Alliot. "Certified global minima for a benchmark of
            difficult optimization problems". arXiv preprint arXiv:2003.09867
            2020.

        """
        super().__init__(d, 'Michalewicz')

        self.par_m = m

        self.set_lim(0., np.pi)

        if self.d == 2:
            self.set_min([2.20, 1.57], -1.8013)
        if self.d == 5:
            self.set_min(None, -4.687658)
        if self.d == 10:
            self.set_min(None, -9.66015)

    def _calc(self, x):
        y1 = np.sin(((np.arange(self.d) + 1) * x**2 / np.pi))

        y = -np.sum(np.sin(x) * y1**(2 * self.par_m))

        return y

    def _comp(self, X):
        y1 = np.sin(((np.arange(self.d) + 1) * X**2 / np.pi))

        y = -np.sum(np.sin(X) * y1**(2 * self.par_m), axis=1)

        return y



"""Package teneva, module demo.demo_func_rastrigin: function.

This module contains class that implements analytical Rastrigin function
for demo and tests.

"""
import numpy as np


from .demo_func import DemoFunc


class DemoFuncRastrigin(DemoFunc):
    def __init__(self, d, A=10.):
        """Rastrigin function for demo and tests.

        Args:
            d (int): number of dimensions.
            A (float): parameter of the function.

        Note:
            See https://www.sfu.ca/~ssurjano/rastr.html for details.

            See also Johannes M Dieterich, Bernd Hartke. "Empirical review of
            standard benchmark functions using evolutionary global
            optimization". Applied Mathematics 2012; 3:1552-1564.

        """
        super().__init__(d, 'Rastrigin')

        self.par_A = A

        self.set_lim(-5.12, +5.12)
        self.set_min([0.]*self.d, 0.)

    def _calc(self, x):
        y1 = self.par_A * self.d
        y2 = np.sum(x**2 - self.par_A * np.cos(2. * np.pi * x))
        return y1 + y2

    def _comp(self, X):
        y1 = self.par_A * self.d
        y2 = np.sum(X**2 - self.par_A * np.cos(2. * np.pi * X), axis=1)
        return y1 + y2



"""Package teneva, module core.tensor: basic operations with TT-tensors.

This module contains the basic operations and utilities for TT-tensors,
including "add", "mul", "norm", "mean", etc.

"""
import numba as nb
import numpy as np


# We use inly matrix_svd here (the whole module is imported to avoid recursion):
import teneva


def accuracy(Y1, Y2):
    """Compute || Y1 - Y2 || / || Y2 || for tensors in the TT-format.

    Args:
        Y1 (list): TT-tensor.
        Y2 (list): TT-tensor.

    Returns:
        float: the relative difference between two tensors.

    """
    return norm(sub(Y1, Y2)) / norm(Y2)


def add(Y1, Y2):
    """Compute Y1 + Y2 in the TT-format.

    Args:
        Y1 (list): TT-tensor (or it may be int/float).
        Y2 (list): TT-tensor (or it may be int/float).

    Returns:
        list: TT-tensor, which represents the element wise sum of Y1 and Y2.
            If both Y1 and Y2 are numbers, then result will be float number.

    """
    if _is_num(Y1) and _is_num(Y2):
        return Y1 + Y2
    elif _is_num(Y1):
        Y1 = const(shape(Y2), Y1)
    elif _is_num(Y2):
        Y2 = const(shape(Y1), Y2)

    n, r1, r2, Y = shape(Y1), ranks(Y1), ranks(Y2), []
    for i, (G1, G2, k) in enumerate(zip(Y1, Y2, n)):
        if i == 0:
            G = np.concatenate([G1, G2], axis=2)
        elif i == len(n) - 1:
            G = np.concatenate([G1, G2], axis=0)
        else:
            r1_l, r1_r = r1[i:i+2]
            r2_l, r2_r = r2[i:i+2]
            Z1 = np.zeros([r1_l, k, r2_r])
            Z2 = np.zeros([r2_l, k, r1_r])
            L1 = np.concatenate([G1, Z1], axis=2)
            L2 = np.concatenate([Z2, G2], axis=2)
            G = np.concatenate([L1, L2], axis=0)
        Y.append(G)

    return Y


def add_many(Y_many, e=1.E-10, r=1.E+12, trunc_freq=15):
    """Compute Y1 + Y2 + ... + Ym in the TT-format.

    Args:
        Y_many (list): the list of TT-tensors (some of them may be int/float).
        e (float): desired approximation accuracy (should be > 0).
        r (int): maximum rank of the result (should be > 0).
        trunc_freq (int): frequency of intermediate summation result truncation.

    Returns:
        list: TT-tensor, which represents the element wise sum of all given
            tensors. If all the tensors are numbers, then result will be
            float number.

    """
    Y = copy(Y_many[0])
    for i, Y_curr in enumerate(Y_many[1:]):
        Y = add(Y, Y_curr)
        if not _is_num(Y) and (i+1) % trunc_freq == 0:
            Y = truncate(Y, e)
    return truncate(Y, e, r) if not _is_num(Y) else Y


def const(n, v=1.):
    """Build tensor with all values equal to given number in the TT-format.

    Args:
        n (list): shape of the tensor.
        v (float): all elements of the tensor will be equal to this value.

    Returns:
        list: TT-tensor with all values equal to the given number "v".

    Note:
        The resulting TT-tensor has all TT-ranks equals to 1.

        TODO: check if we need to distribute value v evenly across TT-cores.

    """
    Y = [np.ones([1, k, 1], dtype=float) for k in n]
    Y[0] *= v
    return Y


def copy(Y):
    """Return a copy of the given TT-tensor.

    Args:
        Y (list): TT-tensor (or it may be int/float).

    Returns:
        list: TT-tensor, which is a copy of the given TT-tensor. If Y is a
            number, then result will be the same number.

    """
    return [G.copy() for G in Y] if not _is_num(Y) else Y


def erank(Y):
    """Compute effective TT-rank of the given TT-tensor.

    Effective TT-rank r of a TT-tensor X with shape [n_1, n_2, ..., n_d] and
    TT-ranks r_0, r_1, ..., r_d (r_0 = r_d = 1) is a solution of quadratic
    equation
    n_1 r + \sum_{\alpha=2}^{d-1} n_\alpha r^2 + n_d r =
    \sum_{\alpha=1}^{d} n_\alpha r_{\alpha-1} r_{\alpha}.

    The representation with a constant TT-rank r (r_0 = 1, r_1 = r_2 = ... =
    r_{d-1} = r, r_d = 1) yields the same total number of parameters as in the
    original decomposition of the tensor X.

    Args:
        Y (list): TT-tensor.

    Returns:
        float: effective TT-rank.

    """
    d, n, r = len(Y), shape(Y), ranks(Y)
    sz = np.dot(n * r[0:d], r[1:])
    b = r[0] * n[0] + n[d-1] * r[d]
    a = np.sum(n[1:d-1])
    return (np.sqrt(b * b + 4 * a * sz) - b) / (2 * a)


def full(Y):
    """Export TT-tensor to the full (numpy) format.

    Args:
        Y (list): TT-tensor.

    Returns:
        np.ndarray: multidimensional array related to the given TT-tensor.

    Note:
         This function can only be used for relatively small tensors. The
         resulting tensor will have n^d elements and may not fit in memory for
         large dimensions.

    """
    Z = Y[0].copy()
    for i in range(1, len(Y)):
        Z = np.tensordot(Z, Y[i], 1)
    return Z[0, ..., 0]


def get(Y, k, to_item=True):
    """Compute the element of the TT-tensor.

    Args:
        Y (list): TT-tensor.
        k (tuple): the multiindex for the tensor.
        to_item (bool): flag, if True, then the float will be returned, and if
            it is False, then the 1-element array will be returned.

    Returns:
        float: the element of the TT-tensor. If "to_item" is False, then
            1-element np.ndarray will be returned.

    """
    Q = Y[0][0, k[0], :]
    for i in range(1, len(Y)):
        Q = np.einsum('q,qp->p', Q, Y[i][:, k[i], :])
    return Q[0] if to_item else Q


def getter(Y, compile=True):
    """Build the fast getter function to compute the element of the TT-tensor.

    Args:
        Y (list): TT-tensor.
        compile (bool): flag, if True, then the getter will be called one time
            with a random multiindex to compile its code.

    Returns:
        function: the function that computes the element of the TT-tensor. It
            has one argument "k" (tuple) which is the multiindex for the tensor.

    Note:
        Note that the gain from using this getter instead of the base function
        "get" appears only in the case of many requests for calculating the
        tensor value (otherwise, the time spent on compiling the getter may
        turn out to be significant).

    """
    Y_nb = tuple([np.array(G, order='C') for G in Y])

    @nb.jit(nopython=True)
    def get(k):
        Q = Y_nb[0]
        y = [Q[0, k[0], r2] for r2 in range(Q.shape[2])]
        for i in range(1, len(Y_nb)):
            Q = Y_nb[i]
            R = np.zeros(Q.shape[2])
            for r1 in range(Q.shape[0]):
                for r2 in range(Q.shape[2]):
                    R[r2] += y[r1] * Q[r1, k[i], r2]
            y = list(R)
        return y[0]

    if compile:
        y = get(np.zeros(len(Y), dtype=int))

    return get


def mean(Y, P=None, norm=True):
    """Compute mean value of the TT-tensor with the given inputs probability.

    Args:
        Y (list): TT-tensor.
        P (list): optional probabilities for each dimension. It is the list of
            length d (number of tensor dimensions), where each element is also
            a list with length equals to the number of tensor elements along the
            related dimension. Hence, P[m][i] relates to the probability of i-th
            input for m-th mode (dimension).
        norm (bool): service (inner) flag, should be True.

    Returns:
        float: the mean value of the TT-tensor.

    """
    R = np.ones((1, 1))
    for i in range(len(Y)):
        k = Y[i].shape[1]
        if P is not None:
            Q = P[i][:k]
        else:
            Q = np.ones(k) / k if norm else np.ones(k)
        R = R @ np.einsum('rmq,m->rq', Y[i], Q)
    return R[0, 0]


def mul(Y1, Y2):
    """Compute element wise product Y1 * Y2 in the TT-format.

    Args:
        Y1 (list): TT-tensor (or it may be int/float).
        Y2 (list): TT-tensor (or it may be int/float).

    Returns:
        list: TT-tensor, which represents the element wise product of Y1 and Y2.
            If both Y1 and Y2 are numbers, then result will be float number.

    """
    if _is_num(Y1) and _is_num(Y2):
        return Y1 * Y2
    elif _is_num(Y1):
        Y = copy(Y2)
        Y[0] *= Y1
        return Y
    elif _is_num(Y2):
        Y = copy(Y1)
        Y[0] *= Y2
        return Y

    Y = []
    for G1, G2 in zip(Y1, Y2):
        G = G1[:, None, :, :, None] * G2[None, :, :, None, :]
        G = G.reshape([G1.shape[0]*G2.shape[0], -1, G1.shape[-1]*G2.shape[-1]])
        Y.append(G)

    return Y


def mul_scalar(Y1, Y2):
    """Compute scalar product for Y1 and Y2 in the TT-format.

    Args:
        Y1 (list): TT-tensor.
        Y2 (list): TT-tensor.

    Returns:
        list: TT-tensor, which represents the scalar product of Y1 and Y2.

    """
    v = None
    for i, (G1, G2) in enumerate(zip(Y1, Y2)):
        G = G1[:, None, :, :, None] * G2[None, :, :, None, :]
        G = G.reshape([G1.shape[0]*G2.shape[0], -1, G1.shape[-1]*G2.shape[-1]])
        G = np.sum(G, axis=1)
        v = G.copy() if i == 0 else v @ G
    return v.item()


def norm(Y):
    """Compute Frobenius norm of the given TT-tensor.

    Args:
        Y (list): TT-tensor.

    Returns:
        float: Frobenius norm of the TT-tensor.

    """
    v = mul_scalar(Y, Y)
    return np.sqrt(v) if v > 1.E-16 else 0.


def orthogonalize(Y, k):
    """Orthogonalize TT-tensor.

    Args:
        Y (list): TT-tensor.
        k (int): desired approximation accuracy (should be > 0).

    Returns:
        list: TT-tensor, which has orthogonalized TT-cores.

    """
    Z = copy(Y)
    L = np.array([[1.]])
    R = np.array([[1.]])

    for i in range(0, k):
        G = _reshape(Z[i], [-1, Z[i].shape[2]])
        Q, R = np.linalg.qr(G, mode='reduced')
        Z[i] = _reshape(Q, Z[i].shape[:-1] + (Q.shape[1], ))
        G = _reshape(Z[i+1], [Z[i+1].shape[0], -1])
        Z[i+1] = _reshape(np.dot(R, G), (R.shape[0], ) + Z[i+1].shape[1:])

    for i in range(len(Z)-1, k, -1):
        G = _reshape(Z[i], [Z[i].shape[0], -1])
        L, Q = scipy.linalg.rq(G, mode='economic', check_finite=False)
        Z[i] = _reshape(Q, (Q.shape[0], ) + Z[i].shape[1:])
        G = _reshape(Z[i-1], [-1, Z[i-1].shape[2]])
        Z[i-1] = _reshape(np.dot(G, L), Z[i-1].shape[:-1] + (L.shape[1], ))

    return Z


def rand(n, r, f=np.random.randn):
    """Construct random TT-tensor.

    Args:
        n (list): shape of the tensor. It should be list or np.ndarray of the
            length d, where d is a number of dimensions.
        r (list): TT-ranks of the tensor. It should be list or np.ndarray of the
            length d+1 with outer elements (first and last) equals to 1. If all
            inner TT-ranks are equal, it may be the int number.
        f (function): sampling function.

    Returns:
        list: TT-tensor.

    """
    n = np.asanyarray(n, dtype=int)
    d = n.size

    if isinstance(r, (int, float)):
        r = [1] + [int(r)] * (d - 1) + [1]
    r = np.asanyarray(r, dtype=int)

    ps = np.cumsum(np.concatenate(([1], n * r[0:d] * r[1:d+1])))
    ps = ps.astype(int)
    core = f(ps[d] - 1)

    Y = []
    for i in range(d):
        G = core[ps[i]-1:ps[i+1]-1]
        Y.append(G.reshape((r[i], n[i], r[i+1]), order='F'))

    return Y


def ranks(Y):
    """Return the TT-ranks of the given TT-tensor.

    Args:
        Y (list): TT-tensor.

    Returns:
        np.ndarray: TT-ranks in form of the 1D array of ints of the length d+1,
            where d is a number of tensor dimensions (the first and last
            elements are equal 1).

    """
    return np.array([1] + [G.shape[2] for G in Y], dtype=int)


def shape(Y):
    """Return the shape of the given TT-tensor.

    Args:
        Y (list): TT-tensor.

    Returns:
        np.ndarray: shape of the tensor in form of the 1D array of ints of the
            length d, where d is a number of tensor dimensions.

    """
    return np.array([G.shape[1] for G in Y], dtype=int)


def show(Y):
    """Display (print) mode sizes and TT-ranks of the given TT-tensor.

    Args:
        Y (list): TT-tensor.

    """
    n, r = shape(Y), ranks(Y)
    l = max(int(np.ceil(np.log10(max(r)+1))) + 1, 3)
    form_str = '{:^' + str(l) + '}'

    s0 = ' '*(l//2)
    s1 = s0 + ''.join([form_str.format(k) for k in n])
    s2 = s0 + ''.join([form_str.format('/ \\') for _ in n])
    s3 = ''.join([form_str.format(q) for q in r])

    print(f'{s1}\n{s2}\n{s3}\n')


def size(Y):
    """Return the size (number of parameters) of the given TT-tensor.

    Args:
        Y (list): TT-tensor.

    Returns:
        int: total number of parameters in the TT-representation (it is a sum
            of sizes of all TT-cores).

    """
    return np.sum([G.size for G in Y])


def sub(Y1, Y2):
    """Compute Y1 - Y2 in the TT-format.

    Args:
        Y1 (list): TT-tensor (or it may be int/float).
        Y2 (list): TT-tensor (or it may be int/float).

    Returns:
        list: TT-tensor, which represents the result of the operation Y1 - Y2.
            If both Y1 and Y2 are numbers, then result will be float number.

    """
    if _is_num(Y1) and _is_num(Y2):
        return Y1 - Y2

    if _is_num(Y2):
        Y2 = const(shape(Y1), -1.*Y2)
    else:
        Y2 = copy(Y2)
        Y2[0] *= -1.

    return add(Y1, Y2)


def sum(Y):
    """Compute sum of all tensor elements.

    Args:
        Y (list): TT-tensor.

    Returns:
        float: the sum of all tensor elements.

    """
    return mean(Y, norm=False)


def truncate(Y, e, r=1.E+12, orth=True):
    """Truncate (round) TT-tensor.

    Args:
        Y (list): TT-tensor wth overestimated ranks.
        e (float): desired approximation accuracy (should be > 0).
        r (int): maximum rank of the result (should be > 0).
        orth (bool): if the flag is set, then tensor orthogonalization will be
            performed.

    Returns:
        list: TT-tensor, which is rounded up to a given accuracy "e" and
            satisfying the rank constraint "r".

    """
    d, n = len(Y), shape(Y)

    if orth:
        Z = orthogonalize(Y, d-1)
        delta = e / np.sqrt(d-1) * np.linalg.norm(Z[-1])
    else:
        Z = copy(Y)
        delta = e

    for k in range(d-1, 0, -1):
        M = _reshape(Z[k], [Z[k].shape[0], -1])
        L, M = teneva.matrix_svd(M, delta, r)
        Z[k] = _reshape(M, [-1, n[k], Z[k].shape[2]])
        Z[k-1] = np.einsum('ijk,kl', Z[k-1], L, optimize=True)

    return Z


def _is_num(A):
    return isinstance(A, (int, float))


def _reshape(A, n):
    return np.reshape(A, n, order='F')
